import sys
from logging import info
from os import makedirs
from os.path import dirname
from collections.abc import Mapping
from configparser import ConfigParser
from ya2.utils.logics import LogicsTools


class DctFile:

    def __init__(self, fpath, default_dct=None, persistent=True):
        default_dct = default_dct or {}
        if sys.platform == 'darwin' and LogicsTools.runtime():
            fpath = dirname(__file__) + '/' + fpath
        self.fpath = fpath
        self.persistent = persistent
        try:
            # with open(fpath) as json: fdct = load(json)
            config = ConfigParser()
            config.read(fpath)
            fdct = {section: dict(config.items(section))
                    for section in config.sections()}
            fdct = self.__typed_dct(fdct)
            self.dct = self.__add_default(default_dct, fdct)
        except IOError:
            self.dct = default_dct

    @staticmethod
    def __typed_dct(dct):
        def convert_single_val(val):
            try:
                return int(val)
            except ValueError:
                try:
                    return float(val)
                except ValueError:
                    if not val or val[0] != '[':
                        return val
                    else:
                        raise ValueError

        def converted(val):
            try:
                return convert_single_val(val)
            except ValueError:
                return [elm.strip() for elm in val[1:-1].split(',')]
        new_dct = {}
        for section, sec_dct in dct.items():
            for key, val in sec_dct.items():
                if section not in new_dct:
                    new_dct[section] = {}
                new_dct[section][key] = converted(val)
        return new_dct

    @staticmethod
    def __add_default(dct, upd):
        for key, val in upd.items():
            if isinstance(val, Mapping):
                dct[key] = DctFile.__add_default(dct.get(key, {}), val)
            else:
                dct[key] = upd[key]
        return dct

    @staticmethod
    def deepupdate(dct, new_dct):
        for key, val in new_dct.items():
            if isinstance(val, Mapping):
                dct[key] = DctFile.deepupdate(dct.get(key, {}), val)
            else:
                dct[key] = val
        return dct

    def store(self):
        info('storing %s' % self.fpath)
        if not self.persistent: return
        # json_str = dumps(self.dct, sort_keys=True, indent=4,
        #                  separators=(',', ': '))
        # with open(self.fpath, 'w') as json: json.write(json_str)
        fdct = {}
        for section, sec_dct in self.dct.items():
            if section not in fdct:
                fdct[section] = {}
            for key, val in sec_dct.items():
                if type(val) == list:
                    fdct[section][key] = '[%s]' % ', '.join(val)
                else:
                    fdct[section][key] = val
        config = ConfigParser()
        for key in self.dct:
            config[key] = fdct[key]
        if dirname(self.fpath):
            makedirs(dirname(self.fpath), exist_ok=True)
        with open(self.fpath, 'w') as ini_file:
            config.write(ini_file)

    def __getitem__(self, arg): return self.dct[arg]

    def __setitem__(self, arg, val): self.dct[arg] = val

    def __delitem__(self, arg): del self.dct[arg]
